"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const DEFAULT_MAX_CHAR_LENGTH = 255;
const DEFAULT_MIN_DATE = '01-01-1970';
const dummyCustomSchema = {
    maxCharLength: DEFAULT_MAX_CHAR_LENGTH,
    ignoredTables: [],
    tables: [],
    values: {}
};
class Analyser {
    constructor(dbConnection, database, customSchema = dummyCustomSchema) {
        this.dbConnection = dbConnection;
        this.database = database;
        this.customSchema = customSchema;
        this.tables = [];
        this.extractTables = () => __awaiter(this, void 0, void 0, function* () {
            const tables = yield this.dbConnection
                .select([
                this.dbConnection.raw('t.TABLE_NAME AS name'),
                this.dbConnection.raw('GROUP_CONCAT(c.REFERENCED_TABLE_NAME SEPARATOR ",") AS referenced_table'),
            ])
                .from('information_schema.tables as t')
                .leftJoin('information_schema.key_column_usage as c', function () {
                this.on('c.CONSTRAINT_SCHEMA', '=', 't.TABLE_SCHEMA')
                    .andOn('c.TABLE_NAME', '=', 't.TABLE_NAME');
            })
                .where('t.TABLE_SCHEMA', this.database)
                .andWhere('t.TABLE_TYPE', 'BASE TABLE')
                .whereNotIn('t.TABLE_NAME', this.customSchema.ignoredTables || [])
                .groupBy('t.TABLE_SCHEMA', 't.TABLE_NAME')
                .orderBy(2);
            for (let t = 0; t < tables.length; t++) {
                const table = tables[t];
                let lines;
                let before;
                let after;
                if (this.customSchema.tables) {
                    const customTable = this.customSchema.tables.find(t => t.name && t.name.toLowerCase() === table.name.toLowerCase());
                    if (customTable) {
                        lines = customTable.lines;
                        before = customTable.before;
                        after = customTable.after;
                    }
                    if (customTable && customTable.columns) {
                        for (const column of customTable.columns) {
                            if (column.foreignKey) {
                                if (table.referenced_table)
                                    table.referenced_table += `,${column.foreignKey.table}`;
                                else
                                    table.referenced_table = `${column.foreignKey.table}`;
                            }
                        }
                    }
                }
                if (lines === undefined)
                    lines = (yield this.dbConnection(table.name).count())[0]['count(*)'];
                table.lines = lines;
                table.before = before;
                table.after = after;
                if (table.referenced_table !== null) {
                    table.referenced_table = table.referenced_table.split(',');
                }
                else {
                    table.referenced_table = [];
                }
            }
            const recursive = (branch) => {
                const table = branch[branch.length - 1];
                while (table.referenced_table.length > 0) {
                    const tableName = table.referenced_table.pop();
                    const referencedTable = tables.find((t) => {
                        return t.name === tableName;
                    });
                    if (referencedTable)
                        recursive([].concat(branch, referencedTable));
                }
                ;
                if (table.referenced_table.length === 0) {
                    if (this.tables.find((t) => t.name.toLowerCase() === table.name.toLowerCase()))
                        return;
                    this.tables.push({
                        name: table.name,
                        lines: table.lines,
                        columns: [],
                        before: table.before,
                        after: table.after,
                    });
                    branch.pop();
                    return;
                }
            };
            tables.forEach((table) => {
                recursive([table]);
            });
        });
        this.extractColumns = () => __awaiter(this, void 0, void 0, function* () {
            var _a;
            for (const table of this.tables) {
                const customTable = Object.assign({
                    name: '',
                    columns: [],
                    lines: 0
                }, (_a = this.customSchema.tables) === null || _a === void 0 ? void 0 : _a.find(t => { var _a; return ((_a = t.name) === null || _a === void 0 ? void 0 : _a.toLowerCase()) === table.name.toLowerCase(); }));
                const columns = yield this.dbConnection.select()
                    .from('information_schema.COLUMNS')
                    .where({
                    'TABLE_SCHEMA': this.database,
                    'TABLE_NAME': table.name
                });
                columns
                    .filter((column) => {
                    return ['enum', 'set'].includes(column.DATA_TYPE || '');
                }).forEach((column) => {
                    column.NUMERIC_PRECISION = column.COLUMN_TYPE.match(/[enum,set]\((.*)\)$/)[1].split('\',\'').length;
                });
                table.columns = columns.map((column) => {
                    const options = {
                        max: 0,
                        min: 0,
                        autoIncrement: false,
                        nullable: false,
                        unique: false,
                        unsigned: false
                    };
                    if (column.IS_NULLABLE === 'YES')
                        options.nullable = true;
                    options.max = column.CHARACTER_MAXIMUM_LENGTH || column.NUMERIC_PRECISION;
                    if (column.COLUMN_TYPE.includes('unsigned'))
                        options.unsigned = true;
                    if (column.EXTRA.includes('auto_increment'))
                        options.autoIncrement = true;
                    if (['timestamp', 'datetime', 'date'].includes(column.DATA_TYPE) && this.customSchema.minDate)
                        options.min = this.customSchema.minDate;
                    return {
                        name: column.COLUMN_NAME,
                        generator: column.DATA_TYPE,
                        options,
                    };
                });
                const subQuery = this.dbConnection
                    .select([
                    'kcu2.table_name',
                    'kcu2.column_name',
                    'kcu2.constraint_schema',
                    this.dbConnection.raw('1 AS unique_index')
                ])
                    .from('information_schema.KEY_COLUMN_USAGE AS kcu2')
                    .innerJoin('information_schema.TABLE_CONSTRAINTS AS tc', function () {
                    this.on('tc.CONSTRAINT_SCHEMA', '=', 'kcu2.CONSTRAINT_SCHEMA')
                        .andOn('tc.TABLE_NAME', '=', 'kcu2.TABLE_NAME')
                        .andOn('tc.CONSTRAINT_NAME', '=', 'kcu2.CONSTRAINT_NAME')
                        .andOnIn('tc.CONSTRAINT_TYPE', ["PRIMARY KEY", "UNIQUE"]);
                })
                    .groupBy(['kcu2.TABLE_NAME', 'kcu2.CONSTRAINT_NAME'])
                    .having(this.dbConnection.raw('count(kcu2.CONSTRAINT_NAME) < 2'))
                    .as('indexes');
                const foreignKeys = yield this.dbConnection.select([
                    'kcu.column_name',
                    'kcu.referenced_table_name',
                    'kcu.referenced_column_name',
                    'unique_index'
                ])
                    .from('information_schema.key_column_usage as kcu')
                    .leftJoin(subQuery, function () {
                    this.on('kcu.table_name', 'indexes.table_name')
                        .andOn('kcu.column_name', 'indexes.column_name')
                        .andOn('kcu.constraint_schema', 'indexes.constraint_schema');
                })
                    .where('kcu.table_name', table.name)
                    .whereNotNull('kcu.referenced_column_name');
                for (let c = 0; c < table.columns.length; c++) {
                    const column = table.columns[c];
                    const customColumn = customTable.columns.find(cc => cc.name.toLowerCase() === column.name.toLowerCase());
                    const match = foreignKeys.find((fk) => fk.column_name.toLowerCase() === column.name.toLowerCase());
                    if (match) {
                        column.foreignKey = { table: match.referenced_table_name, column: match.referenced_column_name };
                        column.options.unique = match.unique_index;
                    }
                    if (customColumn) {
                        column.options = Object.assign({}, column.options, customColumn.options);
                        column.foreignKey = customColumn.foreignKey;
                        column.values = customColumn.values;
                    }
                }
            }
        });
        this.values = customSchema.values || {};
    }
    generateJson() {
        return {
            maxCharLength: this.customSchema.maxCharLength || DEFAULT_MAX_CHAR_LENGTH,
            minDate: this.customSchema.minDate || DEFAULT_MIN_DATE,
            tables: this.tables,
            values: this.values,
        };
    }
}
exports.Analyser = Analyser;
